//
//  simple_depth_float.cpp
//
// Sample Demostrating the use of a floating point depth buffer.
//
// Author: Evan Hart
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <map>

#include <GL/glew.h>
#include <GL/glut.h>

#include <nvGlutWidgets.h>

using std::map;

////////////////////////////////////////////////////////////////////////////////
//
// Globals
//
////////////////////////////////////////////////////////////////////////////////

int width = 600, height = 400;
float distance = -20.0f;
bool closer = false;

//FBO handles
GLuint fbo[3] = { 0, 0, 0};
GLuint depthTex[3] = { 0, 0, 0};
GLuint  colorTex = 0;
GLenum depthFormats[2] = { GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT32F_NV};

int activeFBO = 0;

enum UIOption {
    OPTION_DISPLAY_WIREFRAME,
    OPTION_INVERT_DEPTH_TEST,
    OPTION_USE_MATRIX_FOR_DEPTH_INVERT,
    OPTION_DISPLAY_DEPTH,
    OPTION_ANIMATE,
    OPTION_COUNT
};
bool options[OPTION_COUNT];
map<char, UIOption> optionKeyMap;

nv::GlutUIContext ui;

////////////////////////////////////////////////////////////////////////////////
//
// Functions
//
////////////////////////////////////////////////////////////////////////////////

//
//  keys - keyboard callback function
//
//////////////////////////////////////////////////////////////////////
void keys( unsigned char c, int x, int y) {
    ui.keyboardDown( c, x, y);
    c = tolower(c);

    //if this is an options key, update the associated option
    if (optionKeyMap.find(c) != optionKeyMap.end())
        options[optionKeyMap[c]] = !options[optionKeyMap[c]];

    //handle special options
    switch (c) {
        case '\033':
        case 'q':
            exit(0);
    }

    glutPostRedisplay();
}

//
//  keyup - pass key events to the UI
//
//////////////////////////////////////////////////////////////////////
void keyup(unsigned char c, int x, int y) {
    ui.keyboardDown( c, x, y);
}

//
//  mouse - pass mouse button events to the UI
//
//////////////////////////////////////////////////////////////////////
void mouse( int button, int state, int x, int y) {
    ui.mouse( button, state, glutGetModifiers(), x, y);
}

//
//  motion - pass mouse motion events to the UI
//
//////////////////////////////////////////////////////////////////////
void motion( int x, int y) {
    ui.motion( x, y);
}


//
//  fboMenu - menu callback function, changes FBO config
//
//////////////////////////////////////////////////////////////////////
void fboMenu( int mask) {

    //bottom bit contols the FBO used
    activeFBO = mask & 0x1;

    //third bit controls the depth test sense
    options[OPTION_INVERT_DEPTH_TEST] = (mask & 0x2) != 0;
}

//
//  menu - top-level menu callback function
//
//////////////////////////////////////////////////////////////////////
void menu( int val) {

    keys(val, 0, 0);
}

//
// drawToFBO - render scene to FBO
//
//////////////////////////////////////////////////////////////////////
void drawToFBO() {
    //float FBOs allow extended depth range
    double maxDepth = ( activeFBO == 0) ? 1.0 : 1024.0;

    //bind the active FBO
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo[activeFBO]);

    //When used complimentary depth buffering, the far plane is at 1.0 instead of 0.0
    glClearDepthdNV( options[OPTION_INVERT_DEPTH_TEST] ? 0.0 : maxDepth);

    //Reversing the depth range from z to 1-z operation can be done at viewport scaling
    // This can also be done by scaling Z by -1 in the porjection matrix, but scaling changes trinagle winding as well
    // The matrix method has the advantage of occurring earlier and potentially avoiding rounding
    if (options[OPTION_INVERT_DEPTH_TEST] && !options[OPTION_USE_MATRIX_FOR_DEPTH_INVERT])
        glDepthRangedNV( maxDepth, 0.0);
    else
        glDepthRangedNV( 0.0, maxDepth);

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode( GL_PROJECTION);
    glLoadIdentity();

    if (options[OPTION_INVERT_DEPTH_TEST] && options[OPTION_USE_MATRIX_FOR_DEPTH_INVERT])
        glScalef( 1.0f, 1.0f, -1.0f); //scaling z by -1 after the projection has been applied inverts the Z

    //set a fairly severe projection matrix to amplify the z-fighting 
    gluPerspective( 40.0, (double)width/(double)height, 0.0001, 50.0);

    //translate the plane away by the changing distance
    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef( 0.0f, 0.0f, distance);

    //When we invert the depth, obviously all depth tests must be inverted as well
    glDepthFunc( options[OPTION_INVERT_DEPTH_TEST] ? GL_GEQUAL : GL_LEQUAL);

    //When we invert the depth test by scaling in the matrix, the winding is reversed
    glFrontFace( options[OPTION_INVERT_DEPTH_TEST] && options[OPTION_USE_MATRIX_FOR_DEPTH_INVERT] ? GL_CW : GL_CCW);
    glEnable( GL_DEPTH_TEST);

    if ( options[OPTION_DISPLAY_WIREFRAME])
        glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
    else
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);

    glBegin( GL_TRIANGLES);

    //draw a multi-color triangle with a slight slope
    glColor3f( 0.9f, 0.2f, 0.2f);
    glVertex3f( 0.0f, 7.0f, 1.0f);
    glColor3f( 0.2f, 0.9f, 0.2f);
    glVertex3f( 7.0f, -7.0f, 1.0f);
    glColor3f( 0.2f, 0.2f, 1.0f);
    glVertex3f( -7.0f, -7.0f, 0.95f);

    //draw a solid, flat quad slightly behind the triangle
    glColor3f( 0.9f, 0.9f, 0.2f);
    glVertex3f( -10.0f, -10.0f, 0.0f);
    glVertex3f(  10.0f, -10.0f, 0.0f);
    glVertex3f( -10.0f,  10.0f, 0.0f);
    glVertex3f( -10.0f,  10.0f, 0.0f);
    glVertex3f(  10.0f, -10.0f, 0.0f);
    glVertex3f(  10.0f,  10.0f, 0.0f);
    glEnd();

    //set things back to normal
    glDisable( GL_DEPTH_TEST);
    glClearDepth( 1.0);
    glDepthRange( 0.0, 1.0);
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
    glFrontFace( GL_CCW);

    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0);
}

//
//  doUI - handle all UI processing and rendering
//
//////////////////////////////////////////////////////////////////////
void doUI() {
    nv::Rect nullRect;
    const char * depthFormats[2] = { "24-bit integer", "32-bit floating point"};
    const char * messages[2] = { "Z-Fighting, precision compressed to near plane", "Reduced Z-Fighting, quasi-linear distribution of precision"};

    // start UI processing
    ui.begin();

    ui.setLayout( nv::LAYOUT_NONE, nv::ALIGN_TOP);

    // place one group hung from the top left of the window
    ui.beginGroup();
    ui.setLayout( nv::LAYOUT_VERTICAL, nv::ALIGN_TOP);

    // Place the app name and purpose
    ui.doLabel( nullRect, "Simple Depth Float - maximizing float depth precision");

    // create a horizontal group to place the label and combo side by side
    ui.beginGroup();
    ui.setLayout( nv::LAYOUT_HORIZONTAL, nv::ALIGN_TOP);

    // add a drop list that allows the selection of the depth format
    ui.doLabel( nullRect, "Depth format");
    ui.doComboBox( nullRect, 2, depthFormats, &activeFBO);

    // end the horizontal group
    ui.endGroup();

    // add a checkbox control to select whether the depth is inverted
    ui.doCheckButton( nullRect, "Invert depth test", &options[OPTION_INVERT_DEPTH_TEST]);

    ui.endGroup();
    
    // place a group at the bottom left, showing the status
    ui.beginGroup();
    ui.setLayout( nv::LAYOUT_VERTICAL, nv::ALIGN_BOTTOM);
    
    //Everything, except inverted depths and a flowting point buffer crowd precision to the near plane
    int messageNum = ( activeFBO == 1 && options[OPTION_INVERT_DEPTH_TEST]) ? 1 : 0;
    
    //display the status message
    ui.doLabel( nullRect, messages[messageNum]);
    
    ui.endGroup();

    //end UI processing
    ui.end();
}

//
//  draw - callback function used to display the scene
//
//////////////////////////////////////////////////////////////////////
void draw() {

    //perform the real scene drawing
    drawToFBO();

    glClear( GL_COLOR_BUFFER_BIT);

    glMatrixMode( GL_PROJECTION);
    glLoadIdentity();

    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();
    
    //bind the texture and draw the quad
    glEnable(GL_TEXTURE_2D);

    if (!options[OPTION_DISPLAY_DEPTH])
        glBindTexture( GL_TEXTURE_2D, colorTex);
    else
        glBindTexture( GL_TEXTURE_2D, depthTex[activeFBO]);

    //draw a simple quad to display the scene
    glBegin( GL_QUADS);
    glColor3f( 1.0f, 1.0f, 1.0f);
    glTexCoord2f( 0.0f, 0.0f);
    glVertex3f( -1.0f, -1.0f, 0.0f);
    glTexCoord2f( 1.0f, 0.0f);
    glVertex3f(  1.0f, -1.0f, 0.0f);
    glTexCoord2f( 1.0f, 1.0f);
    glVertex3f(  1.0f,  1.0f, 0.0f);
    glTexCoord2f( 0.0f, 1.0f);
    glVertex3f( -1.0f,  1.0f, 0.0f);
    glEnd();

    //prevent the leakage of state back to the FBO rendering routine
    glDisable(GL_TEXTURE_2D);

    doUI();

    glutSwapBuffers();
}

//
//  idle - callback function used for periodic animation
//
//////////////////////////////////////////////////////////////////////
void idle() {
    static int frame = 0;
    static int msElapsed = 0;
    static float frameTime = 10.0f; //measured in ms, start it as if 100 Hz

    // time over a 10 frame interval
    if ( frame  ==  9) {
        int newTime = glutGet( GLUT_ELAPSED_TIME);
        frameTime = float( newTime - msElapsed) / 10.0f;
        msElapsed = newTime;
        frame = 0;
    }
    else {
        frame++;
    }

    //when animating figure out which direction to go, and whether it is time to reverse directions
    if ( options[OPTION_ANIMATE] ) {

        //move the distance of the objects either closer of further
        if (closer) {
            distance += 0.005f * frameTime;
            if (distance >= -20.0f)
                closer = ! closer;
        }
        else {
            distance -= 0.005f  * frameTime;
            if (distance <= -49.0f)
                closer = ! closer;
        }
    }

    glutPostRedisplay();
}

//
//  createFBOs - create Framebuffer objects for the rendering
//
//////////////////////////////////////////////////////////////////////
void createFBOs() {
    // generate the objects
    glGenTextures( 1, &colorTex);
    glGenTextures( 2, depthTex);
    glGenFramebuffersEXT( 2, fbo);

    //create the simple color texture, shared by all FBOs
    glBindTexture( GL_TEXTURE_2D, colorTex);

    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_FLOAT, NULL);

    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    // generate the different depth formats, and create FBOs all referencing the same color tex for simplicity
    for ( int ii = 0; ii < 2; ii++) {
        glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo[ii]);

        glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, colorTex, 0);

        //create the depth buffer for this FBO, create it as a texture to allow visualization
        glBindTexture( GL_TEXTURE_2D, depthTex[ii]);

        glTexImage2D( GL_TEXTURE_2D, 0, depthFormats[ii], width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);

        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

        glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthTex[ii], 0);

        GLenum status = (GLenum) glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    
        if (status != GL_FRAMEBUFFER_COMPLETE_EXT) {
            glDeleteTextures( 1, &depthTex[ii]);
            depthTex[ii] = 0;
            glDeleteFramebuffersEXT( 1, &fbo[ii]);
            fbo[ii] = 0;
            fprintf( stderr, "Failed to create fbo[%d]\n", ii);
        }
    }

    //reset to safe values
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0);
    glBindTexture( GL_TEXTURE_2D, 0);
}

//
//  destroyFBOs - delete the Framebuffer Objects and associated 
// textures
//////////////////////////////////////////////////////////////////////
void destroyFBOs() {

    glDeleteFramebuffersEXT( 2, fbo);
    glDeleteTextures( 1, &colorTex);
    glDeleteTextures( 2, depthTex);
}

//
//  reshape - call back used when the window resizes
//
//////////////////////////////////////////////////////////////////////
void reshape( int w, int h) {
    ui.reshape( w, h);
    width = w;
    height = h;
    glViewport( 0, 0, w, h);

    //need to recreate FBOs here to match the new window size
    destroyFBOs();
    createFBOs();
}

//
//  init - check extensions and create initial FBO
//
//////////////////////////////////////////////////////////////////////
void init() {
    glewInit();

    if (! glewIsSupported( "GL_VERSION_2_0 GL_EXT_framebuffer_object GL_NV_depth_buffer_float")) {
        fprintf( stderr, "Error: This program requires the following capabilities\n");
        fprintf( stderr, "  OpenGL version 2.0 or later\n");
        fprintf( stderr, "  Support for the GL_EXT_framebuffer_object extension\n");
        fprintf( stderr, "  Support for the GL_NV_depth_buffer_float extension\n\n");
        exit(-1);
    }

    createFBOs();

    //create menus
    int subMenu = glutCreateMenu( fboMenu);
    glutAddMenuEntry( "24-bit depth buffer (normal depth test)", 0x0);
    glutAddMenuEntry( "32-bit floating-point depth buffer (normal depth test)", 0x1);
    glutAddMenuEntry( "24-bit depth buffer (1-z depth test)", 0x2);
    glutAddMenuEntry( "32-bit floating-point depth buffer (1-z depth test)", 0x3);
    glutCreateMenu( menu);
    glutAddSubMenu( "Depth buffer type", subMenu);
    glutAddMenuEntry( "Toggle inverted depth Test    [i]", 'i');
    glutAddMenuEntry( "Toggle wireframe              [w]", 'w');
    glutAddMenuEntry( "Toggle matrix depth inversion [m]", 'm');
    glutAddMenuEntry( "Toggle animation              [ ]", ' ');
    glutAddMenuEntry( "Toggle displaying depth       [d]", 'd');

    glutAttachMenu( GLUT_RIGHT_BUTTON);
}

//
//
//////////////////////////////////////////////////////////////////////
int main( int argc, char **argv) {
    glutInit( &argc, argv);
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize( width, height);

    glutCreateWindow("Simple floating-point depth");

    glutDisplayFunc( draw);
    glutKeyboardFunc( keys);
    glutKeyboardUpFunc( keyup);
    glutMouseFunc( mouse);
    glutMotionFunc( motion);
    glutPassiveMotionFunc( motion);
    glutReshapeFunc( reshape);
    glutIdleFunc( idle);

    // set the animate option to use the space key, and enable it by default
    options[OPTION_ANIMATE] = true;
    optionKeyMap[' '] = OPTION_ANIMATE;

    // set the wireframe option to use the 'w' key, and disable it by default
    options[OPTION_DISPLAY_WIREFRAME] = false;
    optionKeyMap['w'] = OPTION_DISPLAY_WIREFRAME;

    // set the wireframe option to use the 'w' key, and disable it by default
    options[OPTION_DISPLAY_DEPTH] = false;
    optionKeyMap['d'] = OPTION_DISPLAY_DEPTH;

    // set the option to invert depth to use the 'i' key, and disable it by default
    options[OPTION_INVERT_DEPTH_TEST] = false;
    optionKeyMap['i'] = OPTION_INVERT_DEPTH_TEST;

    // set the option to invert depth via the projection matrix to use the 'm' key, and disable it by default
    options[OPTION_USE_MATRIX_FOR_DEPTH_INVERT] = false;
    optionKeyMap['m'] = OPTION_USE_MATRIX_FOR_DEPTH_INVERT;

    //print the help info
    printf( "simple_depth_float - sample showing the usage of GL_NV_depth_buffer_float\n");
    printf( "  Commands:\n");
    printf( "    q / [ESC] - Quit the application\n");
    printf( "    [SPACE]   - Toggle continuous animation\n");
    printf( "    w         - Toggle displaying wireframe\n");
    printf( "    d         - Toggle displaying the depth buffer instead of the color buffer\n");
    printf( "    i         - Toggle using inverted depth\n");
    printf( "    m         - Toggle using the projection matrix to invert depth\n\n");

    init();

    glutMainLoop();

    return 0;
}

